//
// Created by marcin on 06.06.16.
//
#include <jmorecfg.h>
#include <sys/stat.h>
#include "Frame.h"
#include <stdio.h>
#include <stdlib.h>
int radiansToDegrees(double theta);

using namespace std;


int Frame::IMG_SCALE = 2;
int Frame::MAX_POINTS = 20;
int Frame::MIN_FINGER_DEPTH = 1500;
int Frame::MAX_D = 15;
int Frame::MAX_FINGER_ANGLE = 90;
int Frame::MIN_FINGER_ANGLE = 20;

int Frame::MIN_THUMB = 100;
int Frame::MAX_THUMB = 200;

int Frame::MIN_INDEX = 60;
int Frame::MAX_INDEX = 100;

// BGR not RGB!!!!!
//enum Frame::Colors{BLACK,WHITE,RED,GREEN,BLUE,CYAN,MAGENTA,YELLOW};

map<Frame::Colors,Scalar> Frame::colors = {{BLACK,Scalar(0,0,0)},
                                    {WHITE,Scalar(255,255,255)},
                                    {RED,Scalar(0,0,255)},
                                    {GREEN,Scalar(0,255,0)},
                                    {BLUE,Scalar(255,0,0)},
                                    {MAGENTA,Scalar(255,0,255)},
                                    {CYAN,Scalar(255,255,0)},
                                    {YELLOW,Scalar(0,255,255)}
                                            };

#define SAVE_EACH_FRAME 0
#define DEFECTS_COUNT 0
#define AREA_SIZE 0
#define LABEL_TI 0

void Frame::compute(VideoCapture video, int COUNTER) {

    Scalar bounds[2];
    //lower HSV
    bounds[0][0] = 70;
    bounds[0][1] = 80;
    bounds[0][2] = 90;
    //upeer HSV
    bounds[1][0] = 110;
    bounds[1][1] = 250;
    bounds[1][2] = 250;

    video >> original; //capture frame

    Size kSize;
    kSize.height = 3;
    kSize.width = 3;
    double sigma = 0.3*(3/2 - 1) + 0.8;
    GaussianBlur(original,original,kSize,sigma,0.0,4);

//    flip(original,original,1); //flip frame
//
//    pyrDown(original, hsv); //  blur & resize v1
    original.copyTo(hsv);
//    cvResize(im,scaleImg); //resize v2
    cvtColor(hsv, hsv, CV_BGR2HSV); // BGR to HSV colorspace
    blackwhite =  Mat(hsv.rows,hsv.cols,CV_8U);
    inRange(hsv,bounds[0],bounds[1],blackwhite);
    cvtColor(hsv, hsv, CV_HSV2BGR);

    Mat eroded(blackwhite.rows,blackwhite.cols,CV_8UC1);
    cv::erode(blackwhite, eroded, cv::Mat(), cv::Point(-1,-1));


    Mat dilated(eroded.rows,eroded.cols,CV_8UC1);
    dilate(eroded, dilated, cv::Mat(), cv::Point(-1,-1),2);

    Mat tempblackwhite;
    medianBlur(dilated, dilated,7);
    dilated.copyTo(tempblackwhite);
    dilated.copyTo(blackwhite);



    int biggestContourIndex = findBiggestContour(tempblackwhite);
//    drawContours( original, contours,biggestContourIndex, CV_RGB(255,255,255), 2, 8, hierarchy,0, Point() );

    if(biggestContourIndex!=-1){
       cog = extractContourInfo(contours[biggestContourIndex]);
        findFingerTips(biggestContourIndex, COUNTER);

        nameFingers(cog, contourAxisAngle, fingerTips);
    }
}

void Frame::findFingerTips(int bigContour, int COUNTER)
{

    vector<vector<Point>> approxedPolyDP(contours.size());
    vector<vector<Point>> hull(contours.size());
    vector<vector<int>> hullInt(contours.size());
    vector<vector<Vec4i>> defects(contours.size());

    vector<Point> startPts;
    vector<Point> endPts;
    vector<Point> farPts;
    vector<float> depths;

    convexHull(Mat(contours[bigContour]), hull[bigContour], false);
    convexHull(Mat(contours[bigContour]), hullInt[bigContour],false);

    convexityDefects(contours[bigContour],hullInt[bigContour],defects[bigContour]);
//    approxPolyDP( Mat(contours[bigContour]), approxedPolyDP[bigContour],3,true);


    int defectsTotal = defects[bigContour].size();
    #if(DEFECTS_COUNT==1)
        cout<<"Defects count : "<<defectsTotal<<endl;
    #endif


    for (int i = 0; i < defectsTotal; i++) {
        Vec4i & v = (Vec4i &) defects[bigContour].at(i);
        Point ptStart(contours[bigContour][v[0]]);
        startPts.push_back(ptStart);

        Point ptEnd (contours[bigContour][v[1]]);
        endPts.push_back(ptEnd);

        Point ptFar(contours[bigContour][v[2]]);
        farPts.push_back(ptFar);
        depths.push_back( v[3]);


        int r = 0;
        int g = 100;

        int b = 255;

        #if(SAVE_EACH_FRAME == 1)
            cout<<i<<" Defect ("<<ptStart<<","<<ptEnd<<","<<ptFar<<","<<v[3]<<")"<<endl;

            r = abs(255 - i * 5 + 40);
            g = abs(255 - i * 5);
            b = abs(255 - i % 2 + 60);

            imshow("preview",original);
            char filename[100];//= "frame.jpg" % COUNTER;
            char dir[100];//= "frame.jpg" % COUNTER;
            sprintf( dir, "/home/marcin/mgr/project/handtracking/out/frames/%d", COUNTER);
            mkdir(dir,0777);
            sprintf( filename, "/home/marcin/mgr/project/handtracking/out/frames/%d/%d.jpg", COUNTER, i);
            imwrite(filename, original);
            if(waitKey(200) == char('q')) break;
                line(original, ptStart, ptFar, CV_RGB(r, g, b), 2, LINE_AA );
//        line(original, ptEnd, ptFar, CV_RGB(r, g, b), 2,LINE_AA );

//        circle( original, ptStart,   4, Scalar(100,0,255), 2 );
        #endif
//        circle( original, ptFar,   4, colors.at(WHITE), 2 );


    }

    reduceTips(startPts, endPts, farPts, depths);
}

void Frame::reduceTips(vector<Point> startPts, vector<Point> endPts, vector<Point> farPts, vector<float> depths)


{

    int defectsSize = startPts.size();
    vector<int> index;
    fingerTips.clear();
    int fingerCounter = 0;
//    waitKey(200);
    for (int i=0; i < defectsSize; i++) {
        char *str;
        sprintf(str,"%d",i);
        cout<<i<<" finger depth "<<depths[i]<<endl;
//        circle(original,startPts[i], 1, colors.at(RED), 2);
//        circle(original,farPts[i], 1, colors.at(BLUE), 2);
        circle(original,endPts[i], 1, colors.at(GREEN), 2);
        if (depths[i] < MIN_FINGER_DEPTH)    // defect too shallow
            continue;
//        cout<<i<<" finger depth "<<depths[i]<<endl;


//        circle( original, farPts[i],   4, colors.at(BLUE), 2 );
        fingerCounter++;
        int angle = angleBetween(startPts[i], farPts[i], endPts[i]);

        putText(original, str,startPts[i],CV_FONT_HERSHEY_SIMPLEX,1,colors.at(WHITE),1,LINE_4);
        cout<<"points "<< str<<startPts[i]<<farPts[i];
        cout<<endPts[i]<<angle<<" "<<360-angle<<endl;

        if(angle>180){
            angle = 360-angle;
        }

        if( angle>=MAX_FINGER_ANGLE || angle <=MIN_FINGER_ANGLE)
            continue;



//                circle(original, endPts[i], 4, colors.at(BLUE), 2);
//        circle(original, farPts[i], 4, colors.at(GREEN), 2);
        index.push_back(i);

        // this point probably is a finger tip, so add to list
        fingerTips.push_back(endPts[i]);
        int duplicate = false;
        for(Point p : fingerTips){
            int x = (p.x - startPts[i].x)*(p.x - startPts[i].x);
            int y = (p.y - startPts[i].y)*(p.y - startPts[i].y);
//            cout<<x<<" "<<y<<endl;
            double d = sqrt(x+y);
            int maxDist = MAX_D;
//            cout<<startPts[i]<<" "<<d<<endl;
            if(d < maxDist){
                duplicate = true;
                break;
            }

        }
//        circle( original, startPts[i],   4, colors.at(BLACK), 2 );

//            circle(original, startPts[i], 4, colors.at(RED), 2);
//            circle(original, farPts[i], 4, colors.at(BLUE), 2);
//            circle(original, endPts[i], 4, colors.at(GREEN), 2);
        cout<<str<<" "<<duplicate<<endl;
        if(!duplicate)
            fingerTips.push_back(startPts[i]);
    }

//    for(Point p : fingerTips){
////        circle(original, startPts[i], 4, colors.at(RED), 2);
////        circle(original, farPts[i], 4, colors.at(BLUE), 2);
//        cout<<p<<" "<<endl;
//        circle(original,p, MAX_D, colors.at(RED), 2);
//
////        circle(original, p, 4, colors.at(WHITE), 1);
//
//    }

    for(int ix=1;ix<index.size();ix++){
//        waitKey(200);
        Point current = farPts[index[ix]];
        Point prev = farPts[index[ix-1]];
        if(ix==0)
            prev = farPts[index[index.size()-1]];
        Point middle =(current+prev)*.5;
//        Point middle(current.x+(current.x-prev.x)/2,current.y+(current.y-prev.y)/2);
//        circle( original, middle,   4, colors.at(WHITE), 2 );
//        line(original, cog, middle, colors.at(RED), 2, LINE_AA );
//        line(original, middle, startPts[index[ix]], colors.at(RED), 2, LINE_AA );


    }

//    remove_if (fingerTips.begin(),fingerTips.end(), IsOdd);   // 2 4 6 8 ? ? ? ? ?


//    cout << "Finger count : " << fingerTips.size() << endl;
//    if(fingerTips.size() !=5 ){
//        waitKey(100000000);
//    }

    // System.out.println("No. of finger tips: " + fingerTips.size());
}


int Frame::angleBetween(Point start, Point far, Point end)
{
    return
          abs((radiansToDegrees  (
                    atan2(start.y - far.y, start.x - far.x) -
                    atan2(end.y - far.y, end.x - far.x )) ));
}


Point Frame::extractContourInfo(vector<Point> points) {
    Moments moment;
    Point center;
    moment = moments(points, false);
    if (moment.m00 != 0) {   // calculate center
            int xCenter = (int) round(moment.m10/moment.m00);//*IMG_SCALE;
            int yCenter = (int) round(moment.m01/moment.m00);//*IMG_SCALE;

            center.x = xCenter;
            center.y = yCenter;
        }
    this->contourAxisAngle = calculateTilt(moment.m11, moment.m20, moment.m02);


    if (this->fingerTips.size() > 0) {
            int yTotal = 0;
            for(Point pt : this->fingerTips)
                yTotal += pt.y;
            int avgYFinger = yTotal / this->fingerTips.size();
            if (avgYFinger > this->cog.y)   // fingers below COG
                this->contourAxisAngle += 180;
        }
    this->contourAxisAngle = 180 - this->contourAxisAngle;
    return center;
}

int Frame::calculateTilt(double m11, double m20, double m02)
{
    double diff = m20 - m02;
    if (diff == 0) {
        if (m11 == 0)
            return 0;
        else if (m11 > 0)
            return 45;
        else   // m11 < 0
            return -45;
    }

    double theta = 0.5 * atan2(2*m11, diff);
    int tilt = (int) round(radiansToDegrees(theta));

    if ((diff > 0) && (m11 == 0))
        return 0;
    else if ((diff < 0) && (m11 == 0))
        return -90;
    else if ((diff > 0) && (m11 > 0))  // 0 to 45 degrees
        return tilt;
    else if ((diff > 0) && (m11 < 0))  // -45 to 0
        return (180 + tilt);   // change to counter-clockwise angle measure
    else if ((diff < 0) && (m11 > 0))   // 45 to 90
        return tilt;
    else if ((diff < 0) && (m11 < 0))   // -90 to -45
        return (180 + tilt);  // change to counter-clockwise angle measure

    return 0;
}

int radiansToDegrees(double theta) { return theta * 180 / M_PI; }



int Frame::findBiggestContour( Mat &tempblackwhite)  {
    findContours(tempblackwhite, contours,hierarchy, CV_RETR_TREE, CV_CLOCKWISE,Point(0, 0) );
//    findContours(tempblackwhite, contours, CV_RETR_TREE, CV_CLOCKWISE, Point(0, 0) );

    int biggestContourix = -1;
    double areaBiggestContour = 0;
    for(int i=0; i < contours.size(); i++) {
//    while ( !contours.empty()) {
//        if (contours.size() > 0) {
            double area =  contourArea(contours[i]);
            if(area> 1000 && areaBiggestContour<area) {
                #if(AREA==1)
                    cout<<"["<<i<<"/"<<contours.size()<<"] area is = "<<area<<endl;
                #endif
                areaBiggestContour = area;
                biggestContourix = i;

//                return i;
            }
////            CvBox2D box = cvMinAreaRect2(contours);
////            if (box != NULL) {
////                CvSize2D32f size = box.size();
////                float area = size.width() * size.height();
////                if (area > maxArea) {
////                    maxArea = area;
////                    bigContour = contours;
////                }
////            }
//
    }
    return biggestContourix;
}

void Frame::nameFingers(Point cogPt, int contourAxisAngle, vector<Point> fingerTips)
{
    namedFingers.clear();
    for (int i=0; i < fingerTips.size(); i++) {
        FingerName fingerName;
        namedFingers.push_back(fingerName);
    }

    namedFingers =  labelThumbIndex(fingerTips, namedFingers);

    namedFingers = labelUnknowns(namedFingers);
//    printFingers("named fingers", namedFingers);

    // printFingers("revised named fingers", namedFingers);
}

vector<FingerName> Frame::labelThumbIndex(vector<Point> fingerTips,
                                          vector<FingerName> nms)
{
    boolean foundThumb = false;
    boolean foundIndex = false;


    int i = (int) (fingerTips.size() - 1);
    while ((i >= 0)) {
        int angle = angleToCOG(fingerTips[i], cog, contourAxisAngle);
        cout<<"ANGLE "<<angle<<endl;
        if(angle>180){
            angle = 360-angle;
        }
#if(LABEL_TI==1)
        #endif
        if ((angle <=  MAX_THUMB) && (angle > MIN_THUMB) && !foundThumb) {
            nms[i].setType(FingerName::THUMB);
            foundThumb = true;
//            putText(original, "THUMB",fingerTips[i],CV_FONT_HERSHEY_SIMPLEX,0.8,colors.at(WHITE),2,LINE_4);
            cout<<"FOUND THUMB"<<endl;
            #if(LABEL_TI==1)
            circle( original, fingerTips[i],   4, colors.at(MAGENTA), 2 );
            #endif

        }

        if ((angle <= MAX_INDEX) && (angle > MIN_INDEX) && !foundIndex) {
            nms[i].setType(FingerName::INDEX);
            foundIndex = true;
            cout<<"FOUND INDEX"<<endl;
            #if(LABEL_TI==1)
            circle( original, fingerTips[i],   4,colors.at(CYAN), 2 );
                #endif

        }

        if ((angle <= MAX_INDEX) && (angle > MIN_INDEX) && !foundIndex) {
            nms[i].setType(FingerName::INDEX);
            foundIndex = true;
            cout<<"FOUND INDEX"<<endl;
#if(LABEL_TI==1)
            circle( original, fingerTips[i],   4,colors.at(CYAN), 2 );
#endif

        }
        i--;





    }
    return nms;
}


int Frame::angleToCOG(Point tipPt, Point cogPt, int contourAxisAngle)
{
    int yOffset = cogPt.y - tipPt.y;
    int xOffset = tipPt.x - cogPt.x;

    circle( original, cogPt,   4, colors.at(YELLOW), 2 );

    double theta = atan2(xOffset,yOffset);
    int angleTip = (int) round( radiansToDegrees(theta));
    int offsetAngleTip = angleTip + (90 - contourAxisAngle);
    return abs(offsetAngleTip);
}



void Frame::printFingers(String title, vector<FingerName> nms){
    cout<<title<<endl;
    for(FingerName name : nms) {
        cout<<name<<endl;
    }
}

vector<FingerName> Frame::labelUnknowns(vector<FingerName> nms) {
    // find first named finger
    int i = 0;
    while ((i < nms.size()) && (nms[i].getType() == FingerName::UNKNOWN))
        i++;
    if (i == nms.size())
        return;

    FingerName name = nms[i];
    nms =  labelPrev(nms, i, name);
    nms = labelFwd(nms, i, name);

    cout<<"--------"<<endl;
    for(FingerName fn : nms){
//        putText(original, str,,CV_FONT_HERSHEY_SIMPLEX,1,colors.at(WHITE),1,LINE_4);
        cout<<fn<<endl;

    }
    cout<<"--------"<<endl;
    return nms;
}  // end of la


vector<FingerName> Frame::labelPrev(vector<FingerName> nms, int i, FingerName name)
{
    i--;
    while ((i >= 0) && (name.getType() != FingerName::UNKNOWN)){
        if (nms[i].getType() == FingerName::UNKNOWN) {   // unknown finger
           name.getPrev();

            if (!usedName(nms, name))
                nms[i]= name;
        }
        else
            name = nms[i];
        i--;
    }
    return nms;
}

bool Frame::usedName(vector<FingerName> nms, FingerName name)
{
    for(FingerName fn : nms)
        if (fn.getType() == name.getType())
            return true;
    return false;
}


vector<FingerName> Frame::labelFwd(vector<FingerName> nms, int i, FingerName name)
{
    i++;
    while ((i < nms.size()) && (name.getType() != FingerName::UNKNOWN)) {
        if (nms[i].getType() == FingerName::UNKNOWN) {  // unknown finger
            name.getNext();
            if (!usedName(nms, name))
                nms[i]=name;
        }
        else
            name = nms[i];
        i++;
    }
    return nms;
}
